/*-
 * Copyright (c) 2013-2017, Mellanox Technologies, Ltd.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS `AS IS' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <linux/module.h>
#include <dev/mlx5/driver.h>
#include "mlx5_core.h"

static void mlx5_vqos_destroy_scheduling_element(struct mlx5_core_dev *mdev,
						 u32 sched_element_id)
{
	u32 in[MLX5_ST_SZ_DW(destroy_scheduling_element_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(destroy_scheduling_element_out)] = {0};

	MLX5_SET(destroy_scheduling_element_in, in, opcode,
		 MLX5_CMD_OP_DESTROY_SCHEDULING_ELEMENT);

	MLX5_SET(destroy_scheduling_element_in, in, scheduling_hierarchy,
		 MLX5_SCHEDULING_ELEMENT_IN_HIERARCHY_E_SWITCH);
	MLX5_SET(destroy_scheduling_element_in, in, scheduling_element_id,
		 sched_element_id);

	mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
}

int mlx5_vqos_create_arbiter(struct mlx5_core_dev *mdev, int type,
			     u32 parent_id,
			     u32 bw_share, u32  max_average_bw,
			     u32 *arbiter_id)
{
	u32 in[MLX5_ST_SZ_DW(create_scheduling_element_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(create_scheduling_element_out)] = {0};
	int err;
	void *sched_ctx;
	void *tsar_attributes;

	MLX5_SET(create_scheduling_element_in, in, opcode,
		 MLX5_CMD_OP_CREATE_SCHEDULING_ELEMENT);
	MLX5_SET(create_scheduling_element_in, in, scheduling_hierarchy,
		 MLX5_SCHEDULING_ELEMENT_IN_HIERARCHY_E_SWITCH);

	sched_ctx = MLX5_ADDR_OF(create_scheduling_element_in, in,
				 scheduling_context);

	MLX5_SET(scheduling_context, sched_ctx, element_type,
		 MLX5_SCHEDULING_CONTEXT_ELEMENT_TYPE_TSAR);
	MLX5_SET(scheduling_context, sched_ctx, parent_element_id, parent_id);
	MLX5_SET(scheduling_context, sched_ctx, bw_share, bw_share);
	MLX5_SET(scheduling_context, sched_ctx, max_average_bw, max_average_bw);

	tsar_attributes = MLX5_ADDR_OF(scheduling_context, sched_ctx,
				       element_attributes);
	MLX5_SET(tsar_element_attributes, tsar_attributes, tsar_type, type);

	err = mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
	if (err) {
		mlx5_core_warn(mdev, "OP_CREATE_SCHEDLING_ELEM(TSAR) failed\n");
		return err;
	}

	*arbiter_id = MLX5_GET(create_scheduling_element_out, out,
				     scheduling_element_id);
	return 0;
}
EXPORT_SYMBOL_GPL(mlx5_vqos_create_arbiter);

void mlx5_vqos_destroy_arbiter(struct mlx5_core_dev *mdev, u32 arbiter_id)
{
	mlx5_vqos_destroy_scheduling_element(mdev, arbiter_id);
}
EXPORT_SYMBOL_GPL(mlx5_vqos_destroy_arbiter);

int mlx5_vqos_attach_vport(struct mlx5_core_dev *mdev, int type,
			   u32 attributes, u32 parent_tsar_id,
			   u32 *sched_element_id)
{
	u32 in[MLX5_ST_SZ_DW(create_scheduling_element_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(create_scheduling_element_out)] = {0};
	int err;

	MLX5_SET(create_scheduling_element_in, in, opcode,
		 MLX5_CMD_OP_CREATE_SCHEDULING_ELEMENT);
	MLX5_SET(create_scheduling_element_in, in, scheduling_hierarchy,
		 MLX5_SCHEDULING_ELEMENT_IN_HIERARCHY_E_SWITCH);

	MLX5_SET(create_scheduling_element_in, in,
		 scheduling_context.element_type, type);
	MLX5_SET(create_scheduling_element_in, in,
		 scheduling_context.parent_element_id, parent_tsar_id);
	MLX5_SET(create_scheduling_element_in, in,
		 scheduling_context.element_attributes, attributes);

	err = mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
	if (err) {
		mlx5_core_warn(mdev, "OP_CREATE_SCHEDLING_ELEM(VPORT) fail\n");
		return err;
	}

	*sched_element_id = MLX5_GET(create_scheduling_element_out, out,
				     scheduling_element_id);

	return 0;
}
EXPORT_SYMBOL_GPL(mlx5_vqos_attach_vport);

void mlx5_vqos_detach_vport(struct mlx5_core_dev *mdev, u32 sched_element_id)
{
	mlx5_vqos_destroy_arbiter(mdev, sched_element_id);
}
EXPORT_SYMBOL_GPL(mlx5_vqos_detach_vport);

int mlx5_vqos_modify_arbiter(struct mlx5_core_dev *mdev, u32 arbiter_id,
			     u32 bw_share, u32 max_average_bw)
{
	u32 in[MLX5_ST_SZ_DW(modify_scheduling_element_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(modify_scheduling_element_out)] = {0};
	int err;

	MLX5_SET(modify_scheduling_element_in, in, opcode,
		 MLX5_CMD_OP_MODIFY_SCHEDULING_ELEMENT);

	MLX5_SET(modify_scheduling_element_in, in, scheduling_hierarchy,
		 MLX5_SCHEDULING_ELEMENT_IN_HIERARCHY_E_SWITCH);
	MLX5_SET(modify_scheduling_element_in, in, scheduling_element_id,
		 arbiter_id);

	MLX5_SET(modify_scheduling_element_in, in,
		 modify_bitmask,
		 MLX5_MODIFY_SCHEDULING_ELEMENT_BITMASK_BW_SHARE |
		 MLX5_MODIFY_SCHEDULING_ELEMENT_BITMASK_MAX_AVERAGE_BW);

	MLX5_SET(modify_scheduling_element_in, in,
		 scheduling_context.bw_share, bw_share);
	MLX5_SET(modify_scheduling_element_in, in,
		 scheduling_context.max_average_bw, max_average_bw);

	err = mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
	if (err)
		mlx5_core_warn(mdev, "MLX5_CMD_OP_MODIFY_SCHED_ELEMENT fail\n");

	return err;
}
EXPORT_SYMBOL_GPL(mlx5_vqos_modify_arbiter);

int mlx5_create_qos_para_vport(struct mlx5_core_dev *mdev, u16 *vport_id)
{
	u32 in[MLX5_ST_SZ_DW(create_qos_para_vport_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(create_qos_para_vport_out)] = {0};
	int err;

	MLX5_SET(create_qos_para_vport_in,
		 in, opcode, MLX5_CMD_OP_CREATE_QOS_PARA_VPORT);

	err = mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
	if (err) {
		mlx5_core_warn(mdev,
			       "MLX5_CMD_OP_CREATE_QOS_PARA_VPORT failed\n");
		return err;
	}

	*vport_id = MLX5_GET(create_qos_para_vport_out, out,
			     qos_para_vport_number);
	return 0;
}
EXPORT_SYMBOL_GPL(mlx5_create_qos_para_vport);

void mlx5_destroy_qos_para_vport(struct mlx5_core_dev *mdev, u16 vport_id)
{
	u32 in[MLX5_ST_SZ_DW(destroy_qos_para_vport_in)] = {0};
	u32 out[MLX5_ST_SZ_DW(destroy_qos_para_vport_out)] = {0};

	MLX5_SET(destroy_qos_para_vport_in,
		 in, opcode, MLX5_CMD_OP_DESTROY_QOS_PARA_VPORT);

	MLX5_SET(destroy_qos_para_vport_in, in,
		 qos_para_vport_number, vport_id);

	mlx5_cmd_exec(mdev, in, sizeof(in), out, sizeof(out));
}
EXPORT_SYMBOL_GPL(mlx5_destroy_qos_para_vport);
